#include <tfhe/tfhe.h>
#include <tfhe/tfhe_io.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <sys/time.h>

void generate_keys(
    TFheGateBootstrappingParameterSet **params,
    TFheGateBootstrappingSecretKeySet **key)
{
  const int minimum_lambda = 110;
  uint32_t seed[] = { 542, 3594, 620 };
  tfhe_random_generator_setSeed(seed,3);

  *params = new_default_gate_bootstrapping_parameters(minimum_lambda);
  *key = new_random_gate_bootstrapping_secret_keyset(*params);
}

LweSample* encrypt_4bit(
    const uint8_t value,
    const TFheGateBootstrappingParameterSet *params,
    const TFheGateBootstrappingSecretKeySet *key)
{
  LweSample* ctxt = new_gate_bootstrapping_ciphertext_array(4, params);
  if (value > 0xF) {
    fprintf(stderr, "引数が4bitの範囲を超えています\n");
    exit(EXIT_FAILURE);
  }

  for (int i=0; i<4; i++) {
    bootsSymEncrypt(&ctxt[i], (value>>i)&1, key);
  }

  return ctxt;
}

uint8_t decrypt_4bit(
    const LweSample *ctxt,
    const TFheGateBootstrappingSecretKeySet *key)
{
  uint8_t ptxt = 0;
  for (int i=0; i<4; i++) {
    int value = bootsSymDecrypt(&ctxt[i], key);
    ptxt |= (value << i);
  }

  return ptxt;
}

void calc(
    const LweSample *ctxt_inA,
    const LweSample *ctxt_inB,
    LweSample *ctxt_res,
    const TFheGateBootstrappingCloudKeySet *bk)
{
  // --------------------------------------------------------------------
  // ここに4bit加算器を実装してください
  // サンプルコードとして、ビットごとの AND を行う処理が実装されています。
  // --------------------------------------------------------------------
  for(int i=0; i<4; i++){
    bootsAND(&ctxt_res[i], &ctxt_inA[i], &ctxt_inB[i], bk);
  }

}

int main(){
  TFheGateBootstrappingParameterSet *params;
  TFheGateBootstrappingSecretKeySet *key;
  generate_keys(&params, &key);

  // 入力の値。4bitの範囲で自由に書き換えてください。
  uint8_t ptxt_inA = 0b0010;
  uint8_t ptxt_inB = 0b0111;
  LweSample* ctxt_inA = encrypt_4bit(ptxt_inA, params, key);
  LweSample* ctxt_inB = encrypt_4bit(ptxt_inB, params, key);
  LweSample* ctxt_res = encrypt_4bit(0, params, key);


  struct timeval tval_before, tval_after, tval_result;

  gettimeofday(&tval_before, NULL);
  calc(ctxt_inA, ctxt_inB, ctxt_res, &key->cloud);
  gettimeofday(&tval_after, NULL);

  // 最上位ビットのキャリー出力は無視します。
  uint8_t expected = (ptxt_inA + ptxt_inB) & 0xF;

  uint8_t ptxt_res = decrypt_4bit(ctxt_res, key);
  printf("expected:%d actual:%d\n", expected, ptxt_res);
  if (expected == ptxt_res) {
    printf("Answer is correct!\n");
  } else {
    printf("Answer is wrong!\n");
  }

  timersub(&tval_after, &tval_before, &tval_result);
  double elapsed_ms = (tval_result.tv_sec * 1000) + ((double)tval_result.tv_usec)/1000;
  printf("Time elapsed: %lfms\n", elapsed_ms);

  // お片付け
  delete_gate_bootstrapping_ciphertext_array(4, ctxt_inA);
  delete_gate_bootstrapping_ciphertext_array(4, ctxt_inB);
  delete_gate_bootstrapping_ciphertext_array(4, ctxt_res);
  delete_gate_bootstrapping_secret_keyset(key);
  delete_gate_bootstrapping_parameters(params);
  return 0;
}
